@@ -93,7 +93,7 @@ bool read_key_file
 	if (!ifs.eof())
 	{
 		ifs.close();
-		std::cerr << "ERROR: reading private key file until EOF failed" << std::endl;
+		std::cerr << "ERROR: reading public/private key file until EOF failed" << std::endl;
 		return false;
 	}
 	ifs.close();
@@ -101,6 +101,31 @@ bool read_key_file
 	return true;
 }
 
+bool read_binary_key_file
+	(const std::string &filename, const tmcg_byte_t type, std::string &result)
+{
+	// read the public/private key from file and convert to ASCII armor
+	tmcg_octets_t input;
+	std::ifstream ifs(filename.c_str(), std::ifstream::in);
+	if (!ifs.is_open())
+	{
+		std::cerr << "ERROR: cannot open public/private key file" << std::endl;
+		return false;
+	}
+	char c;
+	while (ifs.get(c))
+		input.push_back(c);
+	if (!ifs.eof())
+	{
+		ifs.close();
+		std::cerr << "ERROR: reading public/private key file until EOF failed" << std::endl;
+		return false;
+	}
+	ifs.close();
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(type, input, result);
+	return true;
+}
+
 bool read_message
 	(const std::string &filename, std::string &result)
 {
@@ -126,6 +151,31 @@ bool read_message
 	return true;
 }
 
+bool read_binary_message
+	(const std::string &filename, std::string &result)
+{
+	// read the (encrypted) message from file and convert to ASCII armor
+	tmcg_octets_t input;
+	std::ifstream ifs(filename.c_str(), std::ifstream::in);
+	if (!ifs.is_open())
+	{
+		std::cerr << "ERROR: cannot open input file" << std::endl;
+		return false;
+	}
+	char c;
+	while (ifs.get(c))
+		input.push_back(c);
+	if (!ifs.eof())
+	{
+		ifs.close();
+		std::cerr << "ERROR: reading from input file until EOF failed" << std::endl;
+		return false;
+	}
+	ifs.close();
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(1, input, result);
+	return true;
+}
+
 bool write_message
 	(const std::string &filename, const tmcg_octets_t &msg)
 {
@@ -272,13 +322,13 @@ void cleanup_containers
 bool parse_message
 	(const std::string &in, tmcg_octets_t &enc_out, bool &have_seipd_out)
 {
-	// parse encrypted message
+	// decode ASCII armor and parse encrypted message
 	bool have_pkesk = false, have_sed = false;
 	tmcg_byte_t atype = 0, ptag = 0xFF;
 	tmcg_octets_t pkts;
 	atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
 	if (opt_verbose)
-		std::cout << "ArmorDecode() = " << (int)atype << std::endl;
+		std::cout << "ArmorDecode() = " << (int)atype << " with " << pkts.size() << " bytes" << std::endl;
 	if (atype != 1)
 	{
 		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
@@ -297,7 +347,7 @@ bool parse_message
 			std::cout << "PacketDecode() = " << (int)ptag;
 		if (!ptag)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed" << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << pkts.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
 			return false;
@@ -414,7 +464,7 @@ bool parse_message
 	gcry_mpi_powm(tmp, gk, elg_q, elg_p);
 	if (gcry_mpi_cmp_ui(tmp, 1L))
 	{
-		std::cerr << "ERROR: (g^k)^q \equiv 1 mod p not satisfied" << std::endl;
+		std::cerr << "ERROR: (g^k)^q equiv 1 mod p not satisfied" << std::endl;
 		gcry_mpi_release(tmp);
 		return false;
 	}
@@ -469,7 +519,7 @@ bool decrypt_message
 			std::cout << "PacketDecode() = " << (int)ptag;
 		if (!ptag)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed" << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << litmdc.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
 			return false;
@@ -541,15 +591,16 @@ bool decrypt_message
 }
 
 bool parse_signature
-	(const std::string &in, tmcg_byte_t stype, time_t &sigcreationtime_out, time_t &sigexpirationtime_out, tmcg_byte_t &hashalgo_out, tmcg_octets_t &trailer_out)
+	(const std::string &in, tmcg_byte_t stype,
+	time_t &sigcreationtime_out, time_t &sigexpirationtime_out, tmcg_byte_t &hashalgo_out, tmcg_octets_t &trailer_out, bool &sigV3_out)
 {
-	// parse the signature according to OpenPGP
+	// decode ASCII armor and parse the signature according to OpenPGP
 	bool sig = false;
 	tmcg_byte_t atype = 0, ptag = 0xFF;
 	tmcg_octets_t pkts;
 	atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
 	if (opt_verbose)
-		std::cout << "ArmorDecode() = " << (int)atype << std::endl;
+		std::cout << "ArmorDecode() = " << (int)atype << " with " << pkts.size() << " bytes" << std::endl;
 	if (atype != 2)
 	{
 		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
@@ -565,7 +616,7 @@ bool parse_signature
 		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
 		if (!ptag)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed" << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << pkts.size() << std::endl;
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
 			return false; // parsing error detected
@@ -595,17 +646,33 @@ bool parse_signature
 					time_t kmax = ctx.sigcreationtime + ctx.sigexpirationtime;
 					if (ctx.sigexpirationtime && (time(NULL) > kmax))
 						std::cerr << "WARNING: DSA signature is expired" << std::endl;
-					sig = true;
-					// construct the V4 trailer
+					// construct the trailer
 					trailer_out.clear();
-					trailer_out.push_back(4); // V4 format
-					trailer_out.push_back(ctx.type); // type (e.g. 0x00 Binary Document)
-					trailer_out.push_back(ctx.pkalgo); // public-key algorithm (i.e. DSA)
-					trailer_out.push_back(ctx.hashalgo); // hash algorithm
-					trailer_out.push_back(ctx.hspdlen >> 8); // length of hashed subpacket data
-					trailer_out.push_back(ctx.hspdlen);
-					for (size_t i = 0; i < ctx.hspdlen; i++)
-						trailer_out.push_back(ctx.hspd[i]); // hashed subpacket data
+					if (ctx.version == 3)
+					{
+						tmcg_octets_t sigtime_octets; // V3 format
+						std::cerr << "WARNING: V3 signature packet detected; verification may fail" << std::endl;
+						sig = true;
+						sigV3_out = true;
+						CallasDonnerhackeFinneyShawThayerRFC4880::PacketTimeEncode(ctx.sigcreationtime, sigtime_octets);
+						trailer_out.push_back(ctx.type); // type (e.g. 0x00 Binary Document)
+						trailer_out.insert(trailer_out.end(), sigtime_octets.begin(), sigtime_octets.end()); // creation time
+					}
+					else if (ctx.version == 4)
+					{
+						sig = true;
+						sigV3_out = false;
+						trailer_out.push_back(4); // V4 format
+						trailer_out.push_back(ctx.type); // type (e.g. 0x00 Binary Document)
+						trailer_out.push_back(ctx.pkalgo); // public-key algorithm (i.e. DSA)
+						trailer_out.push_back(ctx.hashalgo); // hash algorithm
+						trailer_out.push_back(ctx.hspdlen >> 8); // length of hashed subpacket data
+						trailer_out.push_back(ctx.hspdlen);
+						for (size_t i = 0; i < ctx.hspdlen; i++)
+							trailer_out.push_back(ctx.hspd[i]); // hashed subpacket data
+					}
+					else
+						std::cerr << "WARNING: unrecognized signature packet version " << (int)ctx.version << std::endl;
 				}
 				else if (sig && (ctx.type == stype) && CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
@@ -624,36 +691,33 @@ bool parse_signature
 }
 
 bool parse_public_key
-	(const std::string &in, time_t &keycreationtime_out, time_t &keyexpirationtime_out)
+	(const std::string &in, time_t &keycreationtime_out, time_t &keyexpirationtime_out, bool elg_required)
 {
+	// decode ASCII Armor
+	tmcg_octets_t pkts;
+	tmcg_byte_t atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
+	if (opt_verbose)
+		std::cout << "ArmorDecode() = " << (int)atype << " with " << pkts.size() << " bytes" << std::endl;
+	if (atype != 6)
+	{
+		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
+		return false;
+	}
 	// parse the public key according to OpenPGP
-	bool pubdsa = false, sigdsa = false, subelg = false, sigelg = false;
-	tmcg_byte_t atype = 0, ptag = 0xFF;
+	bool pubdsa = false, sigdsa = false, sigdsaV3 = false, subelg = false, sigelg = false, sigelgV3 = false, uid = false, uat = false;
+	tmcg_byte_t ptag = 0xFF;
 	tmcg_byte_t dsa_sigtype, dsa_pkalgo, dsa_hashalgo, dsa_keyflags[32], elg_sigtype, elg_pkalgo, elg_hashalgo, elg_keyflags[32];
 	tmcg_byte_t dsa_psa[255], dsa_pha[255], dsa_pca[255], elg_psa[255], elg_pha[255], elg_pca[255];
-	tmcg_octets_t pkts;
 	tmcg_octets_t pub_hashing, sub_hashing, issuer, dsa_hspd, elg_hspd, hash;
-	time_t dsa_creation = 0, elg_creation = 0;
+	time_t dsa_creation = 0, dsa_sigtime = 0, elg_creation = 0, elg_sigtime = 0;
 	gcry_mpi_t dsa_r, dsa_s, elg_r, elg_s;
 	gcry_sexp_t dsakey;
 	gcry_error_t ret;
-	size_t erroff;
+	size_t erroff, pnum = 0;
 	dsa_r = gcry_mpi_new(2048);
 	dsa_s = gcry_mpi_new(2048);
 	elg_r = gcry_mpi_new(2048);
 	elg_s = gcry_mpi_new(2048);
-	atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
-	if (opt_verbose)
-		std::cout << "ArmorDecode() = " << (int)atype << std::endl;
-	if (atype != 6)
-	{
-		std::cerr << "ERROR: wrong type of ASCII Armor found (type = " << (int)atype << ")" << std::endl;
-		gcry_mpi_release(dsa_r);
-		gcry_mpi_release(dsa_s);
-		gcry_mpi_release(elg_r);
-		gcry_mpi_release(elg_s);
-		return false;
-	}
 	while (pkts.size() && ptag)
 	{
 		tmcg_openpgp_packet_ctx ctx;
@@ -662,16 +726,17 @@ bool parse_public_key
 		std::vector<std::string> capl;
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
 		ptag = CallasDonnerhackeFinneyShawThayerRFC4880::PacketDecode(pkts, ctx, current_packet, qual, capl, v_i, c_ik);
+		++pnum;
 		if (!ptag)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed" << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at #" << pnum << " and position " << pkts.size() << std::endl;
 			gcry_mpi_release(dsa_r);
 			gcry_mpi_release(dsa_s);
 			gcry_mpi_release(elg_r);
 			gcry_mpi_release(elg_s);
 			cleanup_ctx(ctx);
 			cleanup_containers(qual, v_i, c_ik);
-			return false; // parsing error detected
+			return false;
 		}
 		switch (ptag)
 		{
@@ -679,9 +744,27 @@ bool parse_public_key
 				issuer.clear();
 				for (size_t i = 0; i < sizeof(ctx.issuer); i++)
 					issuer.push_back(ctx.issuer[i]);
-				if (pubdsa && !subelg && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
+				if (pubdsa && !subelg && !uid && !uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
 					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
+					std::cerr << "ERROR: no uid found for this self-signature" << std::endl;
+					gcry_mpi_release(dsa_r);
+					gcry_mpi_release(dsa_s);
+					gcry_mpi_release(elg_r);
+					gcry_mpi_release(elg_s);
+					cleanup_ctx(ctx);
+					cleanup_containers(qual, v_i, c_ik);
+					return false;
+				}
+				else if (pubdsa && !subelg && uid && !uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
+					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
+				{
+					if (ctx.version == 3)
+					{
+						std::cerr << "WARNING: V3 signature packet detected; verification may fail" << std::endl;
+						sigdsaV3 = true;
+						dsa_sigtime = ctx.sigcreationtime;
+					}
 					if (sigdsa)
 						std::cerr << "WARNING: more than one self-signatures; using last signature to check UID" << std::endl;
 					dsa_sigtype = ctx.type;
@@ -699,8 +782,6 @@ bool parse_public_key
 					dsa_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						dsa_hspd.push_back(ctx.hspd[i]);
-					gcry_mpi_set(dsa_r, ctx.r);
-					gcry_mpi_set(dsa_s, ctx.s);
 					if (dsa_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -712,6 +793,8 @@ bool parse_public_key
 						cleanup_containers(qual, v_i, c_ik);
 						return false;
 					}
+					gcry_mpi_set(dsa_r, ctx.r);
+					gcry_mpi_set(dsa_s, ctx.s);
 					if ((dsa_hashalgo < 8) || (dsa_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)dsa_hashalgo << " used for signatures" << std::endl;
 					time_t kmax = dsa_creation + ctx.keyexpirationtime;
@@ -723,11 +806,20 @@ bool parse_public_key
 					for (size_t i = 0; i < current_packet.size(); i++)
 						uidsig.push_back(current_packet[i]);
 				}
-				else if (pubdsa && subelg && (ctx.type == 0x18) && 
+				else if (pubdsa && !subelg && uid && uat && (ctx.type >= 0x10) && (ctx.type <= 0x13) && 
 					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
-					if (sigelg)
-						std::cerr << "WARNING: more than one subkey binding signature; using last signature" << std::endl;
+					std::cerr << "WARNING: ignore certifying self-signature for a user attribute" << std::endl;
+				}
+				else if (pubdsa && subelg && !sigelg && (ctx.type == 0x18) && 
+					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
+				{
+					if (ctx.version == 3)
+					{
+						std::cerr << "WARNING: V3 signature packet detected; verification may fail" << std::endl;
+						sigelgV3 = true;
+						elg_sigtime = ctx.sigcreationtime;
+					}
 					elg_sigtype = ctx.type;
 					elg_pkalgo = ctx.pkalgo;
 					elg_hashalgo = ctx.hashalgo;
@@ -743,8 +835,6 @@ bool parse_public_key
 					elg_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						elg_hspd.push_back(ctx.hspd[i]);
-					gcry_mpi_set(elg_r, ctx.r);
-					gcry_mpi_set(elg_s, ctx.s);
 					if (elg_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -756,6 +846,8 @@ bool parse_public_key
 						cleanup_containers(qual, v_i, c_ik);
 						return false;
 					}
+					gcry_mpi_set(elg_r, ctx.r);
+					gcry_mpi_set(elg_s, ctx.s);
 					if ((elg_hashalgo < 8) || (elg_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)elg_hashalgo << " used for signatures" << std::endl;
 					time_t kmax = elg_creation + ctx.keyexpirationtime;
@@ -767,6 +859,11 @@ bool parse_public_key
 					for (size_t i = 0; i < current_packet.size(); i++)
 						subsig.push_back(current_packet[i]);
 				}
+				else if (pubdsa && subelg && sigelg && (ctx.type == 0x18) && 
+					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
+				{
+					std::cerr << "WARNING: more than one subkey binding signature; using first signature" << std::endl;
+				}
 				else if (pubdsa && !subelg && (ctx.type == 0x20) && // Key revocation signature 
 					CallasDonnerhackeFinneyShawThayerRFC4880::OctetsCompare(keyid, issuer))
 				{
@@ -796,6 +893,13 @@ bool parse_public_key
 						pub_hashing.push_back(pub[i]);
 					keyid.clear();
 					CallasDonnerhackeFinneyShawThayerRFC4880::KeyidCompute(pub_hashing, keyid);
+					if (opt_verbose)
+					{
+						std::cout << "INFO: Key ID of DSA primary key: " << std::hex;
+						for (size_t i = 0; i < keyid.size(); i++)
+							std::cout << (int)keyid[i] << " ";
+						std::cout << std::dec << std::endl;
+					}
 				}
 				else if ((ctx.pkalgo == 17) && pubdsa)
 				{
@@ -812,22 +916,29 @@ bool parse_public_key
 					std::cerr << "WARNING: public-key algorithm " << (int)ctx.pkalgo << " not supported" << std::endl;
 				break;
 			case 13: // User ID Packet
+				if (uid)
+					std::cerr << "WARNING: more than one uid packet found; using last to verify signature" << std::endl;
+				uid = true;
 				userid = "";
 				for (size_t i = 0; i < sizeof(ctx.uid); i++)
+				{
 					if (ctx.uid[i])
 						userid += ctx.uid[i];
 					else
 						break;
+				}
 				break;
 			case 14: // Public-Subkey Packet
-				if ((ctx.pkalgo == 16) && !subelg)
+				if (ctx.pkalgo == 16)
 				{
-					subelg = true;
+					if (subelg)
+						std::cerr << "WARNING: ElGamal subkey already found; the last one is used" << std::endl;
+					subelg = true, sigelg = false;
 					gcry_mpi_set(elg_p, ctx.p);
 					gcry_mpi_set(elg_g, ctx.g);
 					gcry_mpi_set(elg_y, ctx.y);
 					elg_creation = ctx.keycreationtime;
-					keycreationtime_out = ctx.keycreationtime;
+					keycreationtime_out = ctx.keycreationtime; // FIXME: overwrites creation time of primary key
 					sub.clear();
 					CallasDonnerhackeFinneyShawThayerRFC4880::PacketSubEncode(ctx.keycreationtime, ctx.pkalgo,
 						elg_p, dsa_q, elg_g, elg_y, sub);
@@ -836,11 +947,23 @@ bool parse_public_key
 						sub_hashing.push_back(sub[i]);
 					subkeyid.clear();
 					CallasDonnerhackeFinneyShawThayerRFC4880::KeyidCompute(sub_hashing, subkeyid);
+					if (opt_verbose)
+					{
+						std::cout << "INFO: Key ID of ElGamal subkey: " << std::hex;
+						for (size_t i = 0; i < subkeyid.size(); i++)
+							std::cout << (int)subkeyid[i] << " ";
+						std::cout << std::dec << std::endl;
+					}
 				}
-				else if ((ctx.pkalgo == 16) && subelg)
-					std::cerr << "WARNING: ElGamal subkey already found; the first one is used" << std::endl; 
 				else
-					std::cerr << "WARNING: public-key algorithm " << (int)ctx.pkalgo << " not supported" << std::endl;
+					std::cerr << "WARNING: public-key algorithm " << (int)ctx.pkalgo << " for subkey not supported" << std::endl;
+				break;
+			case 17: // User Attribute Packet
+				std::cerr << "WARNING: user attribute packet found; ignored" << std::endl;
+				uat = true;
+				break;
+			case 0xFE: // unrecognized packet content
+				std::cerr << "WARNING: unrecognized content of packet #" << pnum << " at position " << pkts.size() << std::endl;
 				break;
 		}
 		// cleanup allocated buffers and mpi's
@@ -856,7 +979,7 @@ bool parse_public_key
 		gcry_mpi_release(elg_s);
 		return false;
 	}
-	if (!subelg)
+	if (!subelg && elg_required)
 	{
 		std::cerr << "ERROR: no ElGamal subkey found" << std::endl;
 		gcry_mpi_release(dsa_r);
@@ -874,7 +997,7 @@ bool parse_public_key
 		gcry_mpi_release(elg_s);
 		return false;
 	}
-	if (!sigelg)
+	if ((!sigelg && elg_required) || (subelg && !sigelg))
 	{
 		std::cerr << "ERROR: no self-signature for ElGamal subkey found" << std::endl;
 		gcry_mpi_release(dsa_r);
@@ -885,7 +1008,6 @@ bool parse_public_key
 	}
 	
 	// build keys, check key usage and self-signatures
-	tmcg_octets_t dsa_trailer, elg_trailer, dsa_left, elg_left;
 	ret = gcry_sexp_build(&dsakey, &erroff, "(public-key (dsa (p %M) (q %M) (g %M) (y %M)))", dsa_p, dsa_q, dsa_g, dsa_y);
 	if (ret)
 	{
@@ -904,15 +1026,60 @@ bool parse_public_key
 		else
 			break;
 	}
-	dsa_trailer.push_back(4); // only V4 format supported
-	dsa_trailer.push_back(dsa_sigtype);
-	dsa_trailer.push_back(dsa_pkalgo);
-	dsa_trailer.push_back(dsa_hashalgo);
-	dsa_trailer.push_back(dsa_hspd.size() >> 8); // length of hashed subpacket data
-	dsa_trailer.push_back(dsa_hspd.size());
-	dsa_trailer.insert(dsa_trailer.end(), dsa_hspd.begin(), dsa_hspd.end());
+	if (opt_verbose)
+	{
+		std::cout << "key flags on primary key: ";
+		if ((flags & 0x01) == 0x01)
+			std::cout << "C"; // The key may be used to certify other keys.
+		if ((flags & 0x02) == 0x02)
+			std::cout << "S"; // The key may be used to sign data.
+		if ((flags & 0x04) == 0x04)
+			std::cout << "E"; // The key may be used encrypt communications.
+		if ((flags & 0x08) == 0x08)
+			std::cout << "e"; // The key may be used encrypt storage.
+		if ((flags & 0x10) == 0x10)
+			std::cout << "D"; // The private component of this key may have been split by a secret-sharing mechanism.		
+		if ((flags & 0x20) == 0x20)
+			std::cout << "A"; // The key may be used for authentication.
+		if ((flags & 0x80) == 0x80)
+			std::cout << "M"; // The private component of this key may be in the possession of more than one person.
+		std::cout << std::endl;
+		std::cout << "INFO: userid = " << userid << std::endl;
+		std::cout << "INFO: dsa_sigtype = 0x" << std::hex << (int)dsa_sigtype << std::dec << 
+			" dsa_pkalgo = " << (int)dsa_pkalgo << " dsa_hashalgo = " << (int)dsa_hashalgo << " dsa_hspd.size() = " << dsa_hspd.size() << std::endl;
+	}
+	tmcg_octets_t dsa_trailer, dsa_left;
 	hash.clear();
-	CallasDonnerhackeFinneyShawThayerRFC4880::CertificationHash(pub_hashing, userid, dsa_trailer, dsa_hashalgo, hash, dsa_left);
+	if (sigdsaV3)
+	{
+		tmcg_octets_t dsa_sigtime_octets;
+		CallasDonnerhackeFinneyShawThayerRFC4880::PacketTimeEncode(dsa_sigtime, dsa_sigtime_octets);
+		// The concatenation of the data to be signed, the signature type, and
+		// creation time from the Signature packet (5 additional octets) is
+		// hashed. The resulting hash value is used in the signature algorithm.
+		// The high 16 bits (first two octets) of the hash are included in the
+		// Signature packet to provide a quick test to reject some invalid
+		// signatures.
+		// A V3 signature hashes five octets of the packet body, starting from
+		// the signature type field. This data is the signature type, followed
+		// by the four-octet signature time.
+		dsa_trailer.push_back(dsa_sigtype);
+		dsa_trailer.insert(dsa_trailer.end(), dsa_sigtime_octets.begin(), dsa_sigtime_octets.end());
+		CallasDonnerhackeFinneyShawThayerRFC4880::CertificationHashV3(pub_hashing, userid, dsa_trailer, dsa_hashalgo, hash, dsa_left);
+	}
+	else
+	{
+		dsa_trailer.push_back(4); // only V4 format supported
+		dsa_trailer.push_back(dsa_sigtype);
+		dsa_trailer.push_back(dsa_pkalgo);
+		dsa_trailer.push_back(dsa_hashalgo);
+		dsa_trailer.push_back(dsa_hspd.size() >> 8); // length of hashed subpacket data
+		dsa_trailer.push_back(dsa_hspd.size());
+		dsa_trailer.insert(dsa_trailer.end(), dsa_hspd.begin(), dsa_hspd.end());
+		CallasDonnerhackeFinneyShawThayerRFC4880::CertificationHash(pub_hashing, userid, dsa_trailer, dsa_hashalgo, hash, dsa_left);
+	}
+	if (opt_verbose)
+		std::cout << "INFO: dsa_left = " << std::hex << (int)dsa_left[0] << " " << (int)dsa_left[1] << std::dec << std::endl;
 	ret = CallasDonnerhackeFinneyShawThayerRFC4880::AsymmetricVerifyDSA(hash, dsakey, dsa_r, dsa_s);
 	if (ret)
 	{
@@ -925,48 +1092,91 @@ bool parse_public_key
 		gcry_mpi_release(elg_s);
 		return false;
 	}
-	flags = 0;
-	for (size_t i = 0; i < sizeof(elg_keyflags); i++)
-	{
-		if (elg_keyflags[i])
-			flags = (flags << 8) + elg_keyflags[i];
-		else
-			break;
-	}
-	if ((flags & 0x04) != 0x04)
-	{
-		std::cerr << "ERROR: ElGamal subkey cannot used to encrypt communications" << std::endl;
-		gcry_sexp_release(dsakey);
-		gcry_mpi_release(dsa_r);
-		gcry_mpi_release(dsa_s);
-		gcry_mpi_release(elg_r);
-		gcry_mpi_release(elg_s);
-		return false;
-	}
-	elg_trailer.push_back(4); // only V4 format supported
-	elg_trailer.push_back(elg_sigtype);
-	elg_trailer.push_back(elg_pkalgo);
-	elg_trailer.push_back(elg_hashalgo);
-	elg_trailer.push_back(elg_hspd.size() >> 8); // length of hashed subpacket data
-	elg_trailer.push_back(elg_hspd.size());
-	elg_trailer.insert(elg_trailer.end(), elg_hspd.begin(), elg_hspd.end());
-	hash.clear();
-	CallasDonnerhackeFinneyShawThayerRFC4880::SubkeyBindingHash(pub_hashing, sub_hashing, elg_trailer, elg_hashalgo, hash, elg_left);
-	ret = CallasDonnerhackeFinneyShawThayerRFC4880::AsymmetricVerifyDSA(hash, dsakey, elg_r, elg_s);
-	if (ret)
-	{
-		std::cerr << "ERROR: verification of ElGamal subkey self-signature failed (rc = " << gcry_err_code(ret) << ", str = " <<
-			gcry_strerror(ret) <<")" << std::endl;
-		gcry_sexp_release(dsakey);
-		gcry_mpi_release(dsa_r);
-		gcry_mpi_release(dsa_s);
-		gcry_mpi_release(elg_r);
-		gcry_mpi_release(elg_s);
-		return false;
-	}
-	gcry_sexp_release(dsakey);
 	gcry_mpi_release(dsa_r);
 	gcry_mpi_release(dsa_s);
+	if (elg_required || (subelg && sigelg))
+	{
+		flags = 0;
+		for (size_t i = 0; i < sizeof(elg_keyflags); i++)
+		{
+			if (elg_keyflags[i])
+				flags = (flags << 8) + elg_keyflags[i];
+			else
+				break;
+		}
+		if (opt_verbose)
+		{
+			std::cout << "key flags on subkey: ";
+			if ((flags & 0x01) == 0x01)
+				std::cout << "C"; // The key may be used to certify other keys.
+			if ((flags & 0x02) == 0x02)
+				std::cout << "S"; // The key may be used to sign data.
+			if ((flags & 0x04) == 0x04)
+				std::cout << "E"; // The key may be used encrypt communications.
+			if ((flags & 0x08) == 0x08)
+				std::cout << "e"; // The key may be used encrypt storage.
+			if ((flags & 0x10) == 0x10)
+				std::cout << "D"; // The private component of this key may have been split by a secret-sharing mechanism.		
+			if ((flags & 0x20) == 0x20)
+				std::cout << "A"; // The key may be used for authentication.
+			if ((flags & 0x80) == 0x80)
+				std::cout << "M"; // The private component of this key may be in the possession of more than one person.
+			std::cout << std::endl;
+			std::cout << "INFO: elg_sigtype = 0x" << std::hex << (int)elg_sigtype << std::dec << 
+				" elg_pkalgo = " << (int)elg_pkalgo << " elg_hashalgo = " << (int)elg_hashalgo << " elg_hspd.size() = " << elg_hspd.size() << std::endl;
+		}
+		if (elg_required && ((flags & 0x04) != 0x04))
+		{
+			std::cerr << "ERROR: ElGamal subkey cannot used to encrypt communications" << std::endl;
+			gcry_sexp_release(dsakey);
+			gcry_mpi_release(elg_r);
+			gcry_mpi_release(elg_s);
+			return false;
+		}
+		tmcg_octets_t elg_trailer, elg_left;
+		hash.clear();
+		if (sigelgV3)
+		{
+			tmcg_octets_t elg_sigtime_octets;
+			CallasDonnerhackeFinneyShawThayerRFC4880::PacketTimeEncode(elg_sigtime, elg_sigtime_octets);
+			// The concatenation of the data to be signed, the signature type, and
+			// creation time from the Signature packet (5 additional octets) is
+			// hashed. The resulting hash value is used in the signature algorithm.
+			// The high 16 bits (first two octets) of the hash are included in the
+			// Signature packet to provide a quick test to reject some invalid
+			// signatures.
+			// A V3 signature hashes five octets of the packet body, starting from
+			// the signature type field. This data is the signature type, followed
+			// by the four-octet signature time.
+			elg_trailer.push_back(elg_sigtype);
+			elg_trailer.insert(elg_trailer.end(), elg_sigtime_octets.begin(), elg_sigtime_octets.end());
+			CallasDonnerhackeFinneyShawThayerRFC4880::SubkeyBindingHashV3(pub_hashing, sub_hashing, elg_trailer, elg_hashalgo, hash, elg_left);
+		}
+		else
+		{
+			elg_trailer.push_back(4); // only V4 format supported
+			elg_trailer.push_back(elg_sigtype);
+			elg_trailer.push_back(elg_pkalgo);
+			elg_trailer.push_back(elg_hashalgo);
+			elg_trailer.push_back(elg_hspd.size() >> 8); // length of hashed subpacket data
+			elg_trailer.push_back(elg_hspd.size());
+			elg_trailer.insert(elg_trailer.end(), elg_hspd.begin(), elg_hspd.end());
+			CallasDonnerhackeFinneyShawThayerRFC4880::SubkeyBindingHash(pub_hashing, sub_hashing, elg_trailer, elg_hashalgo, hash, elg_left);
+		}
+		if (opt_verbose)
+			std::cout << "INFO: elg_left = " << std::hex << (int)elg_left[0] << " " << (int)elg_left[1] << std::dec << std::endl;
+		ret = CallasDonnerhackeFinneyShawThayerRFC4880::AsymmetricVerifyDSA(hash, dsakey, elg_r, elg_s);
+		if (ret)
+		{
+			std::cerr << "ERROR: verification of ElGamal subkey self-signature failed (rc = " << gcry_err_code(ret) << ", str = " <<
+				gcry_strerror(ret) <<")" << std::endl;
+			gcry_sexp_release(dsakey);
+			gcry_mpi_release(elg_r);
+			gcry_mpi_release(elg_s);
+			return false;
+		}
+	}
+	gcry_sexp_release(dsakey);
 	gcry_mpi_release(elg_r);
 	gcry_mpi_release(elg_s);
 	return true;
@@ -975,13 +1185,22 @@ bool parse_public_key
 bool parse_private_key
 	(const std::string &in, time_t &keycreationtime_out, time_t &keyexpirationtime_out, std::vector<std::string> &capl_out)
 {
+	// decode ASCII Armor
+	tmcg_octets_t pkts;
+	tmcg_byte_t atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
+	if (opt_verbose)
+		std::cout << "ArmorDecode() = " << (int)atype << " with " << pkts.size() << " bytes" << std::endl;
+	if (atype != 5)
+	{
+		std::cerr << "ERROR: wrong type of ASCII Armor found" << std::endl;
+		exit(-1);
+	}
 	// parse the private key according to OpenPGP
 	bool secdsa = false, sigdsa = false, ssbelg = false, sigelg = false;
-	tmcg_byte_t atype = 0, ptag = 0xFF;
+	tmcg_byte_t ptag = 0xFF;
 	tmcg_byte_t dsa_sigtype, dsa_pkalgo, dsa_hashalgo, dsa_keyflags[32], elg_sigtype, elg_pkalgo, elg_hashalgo, elg_keyflags[32];
 	tmcg_byte_t dsa_psa[255], dsa_pha[255], dsa_pca[255], elg_psa[255], elg_pha[255], elg_pca[255];
 	tmcg_byte_t *key, *iv;
-	tmcg_octets_t pkts;
 	tmcg_octets_t seskey, salt, mpis, hash_input, hash, pub_hashing, sub_hashing, issuer, dsa_hspd, elg_hspd;
 	gcry_cipher_hd_t hd;
 	gcry_error_t ret;
@@ -996,14 +1215,6 @@ bool parse_private_key
 	dsa_s = gcry_mpi_new(2048);
 	elg_r = gcry_mpi_new(2048);
 	elg_s = gcry_mpi_new(2048);
-	atype = CallasDonnerhackeFinneyShawThayerRFC4880::ArmorDecode(in, pkts);
-	if (opt_verbose)
-		std::cout << "ArmorDecode() = " << (int)atype << std::endl;
-	if (atype != 5)
-	{
-		std::cerr << "ERROR: wrong type of ASCII Armor found" << std::endl;
-		exit(-1);
-	}
 	while (pkts.size() && ptag)
 	{
 		tmcg_octets_t current_packet;
@@ -1012,7 +1223,7 @@ bool parse_private_key
 			std::cout << "PacketDecode(pkts.size = " << pkts.size() << ") = " << (int)ptag;
 		if (!ptag)
 		{
-			std::cerr << "ERROR: parsing OpenPGP packets failed" << std::endl;
+			std::cerr << "ERROR: parsing OpenPGP packets failed at position " << pkts.size() << std::endl;
 			gcry_mpi_release(dsa_r);
 			gcry_mpi_release(dsa_s);
 			gcry_mpi_release(elg_r);
@@ -1069,8 +1280,6 @@ bool parse_private_key
 					dsa_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						dsa_hspd.push_back(ctx.hspd[i]);
-					gcry_mpi_set(dsa_r, ctx.r);
-					gcry_mpi_set(dsa_s, ctx.s);
 					if (dsa_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -1082,6 +1291,8 @@ bool parse_private_key
 						cleanup_containers(qual, v_i, x_rvss_qual, c_ik);
 						exit(-1);
 					}
+					gcry_mpi_set(dsa_r, ctx.r);
+					gcry_mpi_set(dsa_s, ctx.s);
 					if ((dsa_hashalgo < 8) || (dsa_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)dsa_hashalgo << 
 							" used for signatures" << std::endl;
@@ -1122,8 +1333,6 @@ bool parse_private_key
 					elg_hspd.clear();
 					for (size_t i = 0; i < ctx.hspdlen; i++)
 						elg_hspd.push_back(ctx.hspd[i]);
-					gcry_mpi_set(elg_r, ctx.r);
-					gcry_mpi_set(elg_s, ctx.s);
 					if (elg_pkalgo != 17)
 					{
 						std::cerr << "ERROR: public-key signature algorithms other than DSA not supported" << std::endl;
@@ -1135,6 +1344,8 @@ bool parse_private_key
 						cleanup_containers(qual, v_i, x_rvss_qual, c_ik);
 						exit(-1);
 					}
+					gcry_mpi_set(elg_r, ctx.r);
+					gcry_mpi_set(elg_s, ctx.s);
 					if ((elg_hashalgo < 8) || (elg_hashalgo >= 11))
 						std::cerr << "WARNING: insecure hash algorithm " << (int)elg_hashalgo << 
 							" used for signatures" << std::endl;
@@ -1251,6 +1462,8 @@ bool parse_private_key
 							if (!mpz_set_gcry_mpi(c_ik[i][k], tmp))
 							{
 								std::cerr << "ERROR: mpz_set_gcry_mpi() failed for tmp" << std::endl;
+								mpz_clear(tmp);
+								delete [] tmp;
 								gcry_mpi_release(dsa_r);
 								gcry_mpi_release(dsa_s);
 								gcry_mpi_release(elg_r);
@@ -1996,6 +2209,8 @@ bool parse_private_key
 						if (!mpz_set_gcry_mpi(v_i[i], tmp))
 						{
 							std::cerr << "ERROR: mpz_set_gcry_mpi() failed for tmp (v_i)" << std::endl;
+							mpz_clear(tmp);
+							delete [] tmp;
 							gcry_mpi_release(dsa_r);
 							gcry_mpi_release(dsa_s);
 							gcry_mpi_release(elg_r);
@@ -2016,6 +2231,8 @@ bool parse_private_key
 							if (!mpz_set_gcry_mpi(c_ik[i][k], tmp))
 							{
 								std::cerr << "ERROR: mpz_set_gcry_mpi() failed for tmp (c_ik)" << std::endl;
+								mpz_clear(tmp);
+								delete [] tmp;
 								gcry_mpi_release(dsa_r);
 								gcry_mpi_release(dsa_s);
 								gcry_mpi_release(elg_r);
@@ -2310,21 +2527,32 @@ bool parse_private_key
 		// cleanup allocated buffers and mpi's
 		cleanup_ctx(ctx);
 		cleanup_containers(qual, v_i, x_rvss_qual, c_ik);
-
 	}
 	if (!secdsa)
 	{
 		std::cerr << "ERROR: no tDSS/DSA private key found" << std::endl;
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 	if (!sigdsa)
 	{
 		std::cerr << "ERROR: no self-signature for tDSS/DSA key found" << std::endl;
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 	if (ssbelg && !sigelg)
 	{
 		std::cerr << "ERROR: no self-signature for ElGamal subkey found" << std::endl;
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 
@@ -2337,6 +2565,10 @@ bool parse_private_key
 	if (ret)
 	{
 		std::cerr << "ERROR: parsing tDSS/DSA key material failed" << std::endl;
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 	size_t flags = 0;
@@ -2369,6 +2601,11 @@ bool parse_private_key
 	if ((flags & 0x02) != 0x02)
 	{
 		std::cerr << "ERROR: tDSS/DSA primary key cannot used to sign data" << std::endl;
+		gcry_sexp_release(dsakey);
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 	dsa_trailer.push_back(4); // only V4 format supported
@@ -2385,6 +2622,11 @@ bool parse_private_key
 	{
 		std::cerr << "ERROR: verification of tDSS/DSA key self-signature failed (rc = " << gcry_err_code(ret) << ", str = " <<
 			gcry_strerror(ret) << ")" << std::endl;
+		gcry_sexp_release(dsakey);
+		gcry_mpi_release(dsa_r);
+		gcry_mpi_release(dsa_s);
+		gcry_mpi_release(elg_r);
+		gcry_mpi_release(elg_s);
 		exit(-1);
 	}
 	if (ssbelg)
@@ -2419,6 +2661,11 @@ bool parse_private_key
 		if ((flags & 0x04) != 0x04)
 		{
 			std::cerr << "ERROR: ElGamal subkey cannot used to encrypt communications" << std::endl;
+			gcry_sexp_release(dsakey);
+			gcry_mpi_release(dsa_r);
+			gcry_mpi_release(dsa_s);
+			gcry_mpi_release(elg_r);
+			gcry_mpi_release(elg_s);
 			exit(-1);
 		}
 		elg_trailer.push_back(4); // only V4 format supported
@@ -2435,6 +2682,11 @@ bool parse_private_key
 		{
 			std::cerr << "ERROR: verification of ElGamal subkey self-signature failed (rc = " << gcry_err_code(ret) << ", str = " <<
 				gcry_strerror(ret) << ")" << std::endl;
+			gcry_sexp_release(dsakey);
+			gcry_mpi_release(dsa_r);
+			gcry_mpi_release(dsa_s);
+			gcry_mpi_release(elg_r);
+			gcry_mpi_release(elg_s);
 			exit(-1);
 		}
 	}
